//////////////////////////////////////////////////////////////////////////////
//
// Minimizer.h,
//
//
//////////////////////////////////////////////////////////////////////////////

#ifndef MINIMIZER_USED
#define MINIMIZER_USED

//////////////////////////////////////////////////////////////////////////////

#include <assert.h>
#include <stdio.h>
#include <cmath>
#include <stdlib.h>
#include <iostream>
using namespace std;

#include "../CC_Mesh/Mesh.h"
#include "DynamicCompCol_Mat_double.h"
#include "compcol_double.h"
#include "mvblasd.h"                     // MV_Vector level 1 BLAS
#include "diagpre_double.h"              // Diagonal preconditioner
#include "cg.h"                          // IML++ CG template
#include "cgs.h"                         // IML++ CGS template
#include "bicg.h"                        // IML++ CGS template

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#define ITMAX 100
#define ZEPS  1.0e-15
#define CGOLD 0.3819660
#define MOV3(a,b,c,d,e,f) (a)=(d);(b)=(e);(c)=(f);
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

class Minimizer
{
 protected:
  int numpoints;
  Mesh *mesh;

//////////////////////////////////////////////////////////////////////////////
// functions that work on all points (seen as vectors in R^3num_points)
  void   add(double *A, double *B, double *C);
  void   sub(double *A, double *B, double *C);
  void   neg(double *A);
  double dot(double *A, double *B);
  void   scale(double s, double *A);
  void   scale_add(double *A, double t, double *B, double *C);
  void   scale_add(double a, double *A, double b, double *B, double *C);
  void   copy(double *A, double *B);
  void   ncopy(double *A, double *B);
  void   set(double *A, double val);

//////////////////////////////////////////////////////////////////////////////

 public:

  inline double                     Energy(double *X);
  inline void                       gradEnergy(double *X, double *g);
  inline DynamicCompCol_Mat_double* HessEnergy(double *X);

//////////////////////////////////////////////////////////////////////////////

  virtual void minimize(Mesh *M)=0;

  Minimizer();
  virtual ~Minimizer();
};

//////////////////////////////////////////////////////////////////////////////

#define MINIMIZE_BENDING_ENERGY 1

double Minimizer::Energy(double *X)
{
#ifdef MINIMIZE_BENDING_ENERGY
  return(mesh->KpEnergy(X));
#else
  return(mesh->KmEnergy(X));
#endif
}

//////////////////////////////////////////////////////////////////////////////

void Minimizer::gradEnergy(double *X, double *g)
{
#ifdef MINIMIZE_BENDING_ENERGY
  mesh->grad_KpEnergy(X, g);
#else
  mesh->grad_KmEnergy(X, g);
#endif
}

//////////////////////////////////////////////////////////////////////////////

DynamicCompCol_Mat_double* Minimizer::HessEnergy(double *X)
{
#ifdef MINIMIZE_BENDING_ENERGY
  return(mesh->Hess_KpEnergy(X));
#else
  return(mesh->Hess_KmEnergy(X));
#endif
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

class CG_Minimizer:public Minimizer
{
  double *DEP;
  double *A, *DEA;
  double *B, *DEB;
  double *G;
  double *H;
  double *tmp;
  double *tmp2;

  double EP, EA, EB;
  double gamma;
  
  double line_Energy(double *MP, double *D, double x);
  double diff_line_Energy(double *MP, double *D, double x);

  double  brent(double *MP, double *D,
		double ax, double bx, double cx, double tol);
  double dbrent(double *MP, double *D,
		double ax, double bx, double cx, double tol);

  double line_minimize(double *, double *);
  void   conjugate_gradient_minimize(void);

 public:

  virtual void minimize(Mesh *M);

  virtual ~CG_Minimizer();
  CG_Minimizer();

};

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

class Newton_Minimizer:public Minimizer
{
  double *P_old;
  double *P_new;
  double *P_tmp;

  // void NewtonStep(Mesh *M);

 public:

  virtual void minimize(Mesh *M);

  virtual ~Newton_Minimizer();
  Newton_Minimizer();
};

//////////////////////////////////////////////////////////////////////////////
 
class CG_Steihaug_Minimizer:public Minimizer
{
  double Delta_bar, eta;

  // mimize()
  double *X, *X_new, *p; //B,g used too

  // CG_Steihaug_direction()
  DynamicCompCol_Mat_double *B;
  double *p1, *p2, *r, *g, *d, *Bd;

  // ModelEnergy
  double *Bp;
  inline double ModelEnergy(DynamicCompCol_Mat_double *B, 
			    double *grad, double *X, double *p);

  int CG_Steihaug_direction(DynamicCompCol_Mat_double *B, 
			    double *grad, double *x_old, 
			    double *p, double Delta, double eps);

 public:

  virtual void minimize(Mesh *M);

  virtual ~CG_Steihaug_Minimizer();
  CG_Steihaug_Minimizer(double Db, double et);

};

//////////////////////////////////////////////////////////////////////////////
// ***************************************************************************
// ***************************************************************************
// ***************************************************************************
double CG_Steihaug_Minimizer::ModelEnergy(DynamicCompCol_Mat_double *B, 
					   double *grad, double *X, double *p)
{
  double f,gp,pBp;

  f  = Energy(X);
  gp = dot(grad,p);
  B->mult(p,Bp);
  pBp= dot(p,Bp);

  return(f+gp+0.5*pBp);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
 
class TrustRegion_Minimizer:public Minimizer
{

 public:

  virtual void minimize(Mesh *M);

  virtual ~TrustRegion_Minimizer();
  TrustRegion_Minimizer();
};

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#endif
