//////////////////////////////////////////////////////////////////////////////
//
// BicubicPatch.C, 07/10/01, Ilja Friedel
//
//
//////////////////////////////////////////////////////////////////////////////

#include "BicubicPatch.h"

//////////////////////////////////////////////////////////////////////////////

int    BicubicPatch::initialized=0;
double BicubicPatch::eX_tensor[256];
double BicubicPatch::iX_tensor[256];
double BicubicPatch::eX_patrick[256];
double BicubicPatch::iX_patrick[256];
double BicubicPatch::Lambda[256];

//////////////////////////////////////////////////////////////////////////////

BicubicPatch::~BicubicPatch()
{
  if(p[0])
    {
      delete[] p[0];
      if(p[0]!=p[1]) delete[] p[1];
      if(p[0]!=p[2]) delete[] p[2];
      p[0]=NULL;
    }
  if(A[0])
    {
      delete[] A[0];
      if(A[0]!=A[1]) delete[] A[1];
      if(A[0]!=A[2]) delete[] A[2];
      A[0]=NULL;
    }
}

//////////////////////////////////////////////////////////////////////////////
// only one dimension
// ordering of points as tensor product
//
// p00 p01 p02 p03
// p04 p05 p06 p07
// p08 p09 p10 p11
// p12 p13 p14 p15

BicubicPatch::BicubicPatch(double *data)
{
  int i,j,k;

  if(!initialized)
    {
      set_matrices();
      initialized=1;
    }

  p[0]=new double[16];
  A[0]=new double[16];

  p[1]=p[0];
  p[2]=p[0];

  A[1]=A[0];
  A[2]=A[0];

  i=0;
      
  for(j=0;j<16;j++)
    {
      p[i][j]=data[j];
      A[i][j]=0.0;
      
      for(k=0;k<16;k++) A[i][j]+=iX_tensor[16*j+k]*data[k];
    }
}

//////////////////////////////////////////////////////////////////////////////
// exactly three dimensions
// ordering of points as defined by Patrick
//

BicubicPatch::BicubicPatch(double **data):Patch()
{
  int i,j,k;

  if(!initialized)
    {
      set_matrices();
      initialized=1;
    }

  for(i=0;i<3;i++)
    {
      p[i]=new double[16];
      A[i]=new double[16];

      for(j=0;j<16;j++)
	{
	  p[i][j]=data[i][j];
	  A[i][j]=0.0;
	  
	  for(k=0;k<16;k++) A[i][j]+=iX_patrick[16*j+k]*data[i][k];
	}
    }
}

//////////////////////////////////////////////////////////////////////////////

void BicubicPatch::print(void)
{
  int i,j;
  
  cout << "BicubicPatch:" <<endl;

  for(i=0;i<3;i++)
    {
      for(j=0;j<16;j++)
	{
	  if(!(j%4)) cout << endl;
	  cout << p[i][j] << "  ";
	}
      cout << endl;
    }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

double BicubicPatch::evaluate_S(const double u, const double v, const int s)
{
  double *a;
  a=A[s];

  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  return(a[0]+(a[ 1]+(a[ 3]+a[ 6]*v)*v)*v+
	(a[2]+(a[ 4]+(a[ 7]+a[10]*v)*v)*v+
	(a[5]+(a[ 8]+(a[11]+a[13]*v)*v)*v+
	(a[9]+(a[12]+(a[14]+a[15]*v)*v)*v)*u)*u)*u);
}

double BicubicPatch::evaluate_D1S(const double u, const double v, const int s)
{
  double *a;
  a=A[s];

  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  return(    a[2]+(    a[ 4]+(    a[ 7]+    a[10]*v)*v)*v+
	(2.0*a[5]+(2.0*a[ 8]+(2.0*a[11]+2.0*a[13]*v)*v)*v+
	(3.0*a[9]+(3.0*a[12]+(3.0*a[14]+3.0*a[15]*v)*v)*v)*u)*u);
}

double BicubicPatch::evaluate_D2S(const double u, const double v, const int s)
{
  double *a;
  a=A[s];

  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  return(a[ 1]+(2.0*a[ 3]+3.0*a[ 6]*v)*v+
	(a[ 4]+(2.0*a[ 7]+3.0*a[10]*v)*v+
	(a[ 8]+(2.0*a[11]+3.0*a[13]*v)*v+
	(a[12]+(2.0*a[14]+3.0*a[15]*v)*v)*u)*u)*u);
}

double BicubicPatch::evaluate_D1D1S(const double u, const double v, const int s)
{
  double *a;
  a=A[s];

  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  return(2.0*a[5]+(2.0*a[ 8]+(2.0*a[11]+2.0*a[13]*v)*v)*v+
	(6.0*a[9]+(6.0*a[12]+(6.0*a[14]+6.0*a[15]*v)*v)*v)*u);
}

double BicubicPatch::evaluate_D1D2S(const double u, const double v, const int s)
{
  double *a;
  a=A[s];

  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  return(    a[ 4]+(2.0*a[ 7]+3.0*a[10]*v)*v+
	(2.0*a[ 8]+(4.0*a[11]+6.0*a[13]*v)*v+
	(3.0*a[12]+(6.0*a[14]+9.0*a[15]*v)*v)*u)*u);
}

double BicubicPatch::evaluate_D2D1S(const double u, const double v, const int s)
{
  double *a;
  a=A[s];

  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  return(    a[ 4]+(2.0*a[ 7]+3.0*a[10]*v)*v+
	(2.0*a[ 8]+(4.0*a[11]+6.0*a[13]*v)*v+
	(3.0*a[12]+(6.0*a[14]+9.0*a[15]*v)*v)*u)*u);
}

double BicubicPatch::evaluate_D2D2S(const double u, const double v, const int s)
{
  double *a;
  a=A[s];

  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  return(2.0*a[ 3]+6.0*a[ 6]*v+
	(2.0*a[ 7]+6.0*a[10]*v+
	(2.0*a[11]+6.0*a[13]*v+
	(2.0*a[14]+6.0*a[15]*v)*u)*u)*u);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

double BicubicPatch::Kp_function(const double u, const double v)
{
  double tmp;
  tmp=u+v;
  exit(1);
}

//////////////////////////////////////////////////////////////////////////////

double BicubicPatch::Km_function(const double u, const double v)
{
  double tmp;
  tmp=u+v;
  exit(1);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void BicubicPatch::add_sample_to_Kp(const double u, const double v, const double weight, double *K)
{
  double g11,g12,g22;
  double gg11,gg12,gg22;
  double Gamma1_11, Gamma1_12, Gamma1_22;
  double Gamma2_11, Gamma2_12, Gamma2_22;
  double det_IS;

  double d1h [16];
  double d2h [16];
  double d11h[16];
  double d12h[16];
  double d22h[16];

  double d1S1, d1S2, d1S3;
  double d2S1, d2S2, d2S3;

  double d11S1, d11S2, d11S3;
  double d12S1, d12S2, d12S3;
  double d22S1, d22S2, d22S3;

  double DomegaS;

  int i,j;

  // setting first partial derivatives of surface

  d1S1  = evaluate_D1S(u,v,0);
  d1S2  = evaluate_D1S(u,v,1);
  d1S3  = evaluate_D1S(u,v,2);

  d2S1  = evaluate_D2S(u,v,0);
  d2S2  = evaluate_D2S(u,v,1);
  d2S3  = evaluate_D2S(u,v,2);

  // setting second partial derivatives of surface

  d11S1 = evaluate_D1D1S(u,v,0);
  d11S2 = evaluate_D1D1S(u,v,1);
  d11S3 = evaluate_D1D1S(u,v,2);

  d12S1 = evaluate_D1D2S(u,v,0);
  d12S2 = evaluate_D1D2S(u,v,1);
  d12S3 = evaluate_D1D2S(u,v,2);

  d22S1 = evaluate_D2D2S(u,v,0);
  d22S2 = evaluate_D2D2S(u,v,1);
  d22S3 = evaluate_D2D2S(u,v,2);

  // setting first derivatives of monomial/eigen basis

  d1h[ 0] = 0.0;
  d1h[ 1] = 0.0;
  d1h[ 2] = 1.0;
  d1h[ 3] = 0.0;
  d1h[ 4] = v;
  d1h[ 5] = 2.0*u;
  d1h[ 6] = 0.0;
  d1h[ 7] = v*v;
  d1h[ 8] = 2.0*u*v;
  d1h[ 9] = 3.0*u*u;
  d1h[10] = v*v*v;
  d1h[11] = 2.0*u*v*v;
  d1h[12] = 3.0*u*u*v;
  d1h[13] = 2.0*u*v*v*v;
  d1h[14] = 3.0*u*u*v*v;
  d1h[15] = 3.0*u*u*v*v*v;

  d2h[ 0] = 0.0;
  d2h[ 1] = 1.0;
  d2h[ 2] = 0.0;
  d2h[ 3] = 2.0*v;
  d2h[ 4] = u;
  d2h[ 5] = 0.0;
  d2h[ 6] = 3.0*v*v;
  d2h[ 7] = 2.0*u*v;
  d2h[ 8] = u*u;
  d2h[ 9] = 0.0;
  d2h[10] = 3.0*u*v*v;
  d2h[11] = 2.0*u*u*v;
  d2h[12] = u*u*u;
  d2h[13] = 3.0*u*u*v*v;
  d2h[14] = 2.0*u*u*u*v;
  d2h[15] = 3.0*u*u*u*v*v;

  // setting second derivatives of monomial/eigen basis

  d11h[ 0] = 0.0;
  d11h[ 1] = 0.0;
  d11h[ 2] = 0.0;
  d11h[ 3] = 0.0;
  d11h[ 4] = 0.0;
  d11h[ 5] = 2.0;
  d11h[ 6] = 0.0;
  d11h[ 7] = 0.0;
  d11h[ 8] = 2.0*v;
  d11h[ 9] = 6.0*u;
  d11h[10] = 0.0;
  d11h[11] = 2.0*v*v;
  d11h[12] = 6.0*u*v;
  d11h[13] = 2.0*v*v*v;
  d11h[14] = 6.0*u*v*v;
  d11h[15] = 6.0*u*v*v*v;

  d12h[ 0] = 0.0;
  d12h[ 1] = 0.0;
  d12h[ 2] = 0.0;
  d12h[ 3] = 0.0;
  d12h[ 4] = 1.0;
  d12h[ 5] = 0.0;
  d12h[ 6] = 0.0;
  d12h[ 7] = 2.0*v;
  d12h[ 8] = 2.0*u;
  d12h[ 9] = 0.0;
  d12h[10] = 3.0*v*v;
  d12h[11] = 4.0*u*v;
  d12h[12] = 3.0*u*u;
  d12h[13] = 6.0*u*v*v;
  d12h[14] = 6.0*u*u*v;
  d12h[15] = 9.0*u*u*v*v;

  d22h[ 0] = 0.0;
  d22h[ 1] = 0.0;
  d22h[ 2] = 0.0;
  d22h[ 3] = 2.0;
  d22h[ 4] = 0.0;
  d22h[ 5] = 0.0;
  d22h[ 6] = 6.0*v;
  d22h[ 7] = 2.0*u;
  d22h[ 8] = 0.0;
  d22h[ 9] = 0.0;
  d22h[10] = 6.0*u*v;
  d22h[11] = 2.0*u*u;
  d22h[12] = 0.0;
  d22h[13] = 6.0*u*u*v;
  d22h[14] = 2.0*u*u*u;
  d22h[15] = 6.0*u*u*u*v;
  
  // setting inverse of first fundamental form IS (times determinant)

  g11 =   d2S1*d2S1 + d2S2*d2S2 + d2S3*d2S3;
  g12 = -(d1S1*d2S1 + d1S2*d2S2 + d1S3*d2S3);
  g22 =   d1S1*d1S1 + d1S2*d1S2 + d1S3*d1S3;

  // determinant of first fundamental form

  det_IS = g11*g22 - g12*g12;

  // setting inverse iIS of first fundamental form IS

  g11 = g11/det_IS;
  g12 = g12/det_IS;
  g22 = g22/det_IS;

  // Christoffel symbols

  Gamma1_11 = g11*(d11S1*d1S1 + d11S2*d1S2 + d11S3*d1S3) + 
              g12*(d11S1*d2S1 + d11S2*d2S2 + d11S3*d2S3);
  Gamma1_12 = g11*(d12S1*d1S1 + d12S2*d1S2 + d12S3*d1S3) + 
              g12*(d12S1*d2S1 + d12S2*d2S2 + d12S3*d2S3);
  Gamma1_22 = g11*(d22S1*d1S1 + d22S2*d1S2 + d22S3*d1S3) + 
              g12*(d22S1*d2S1 + d22S2*d2S2 + d22S3*d2S3);

  Gamma2_11 = g12*(d11S1*d1S1 + d11S2*d1S2 + d11S3*d1S3) + 
              g22*(d11S1*d2S1 + d11S2*d2S2 + d11S3*d2S3);
  Gamma2_12 = g12*(d12S1*d1S1 + d12S2*d1S2 + d12S3*d1S3) + 
              g22*(d12S1*d2S1 + d12S2*d2S2 + d12S3*d2S3);
  Gamma2_22 = g12*(d22S1*d1S1 + d22S2*d1S2 + d22S3*d1S3) + 
              g22*(d22S1*d2S1 + d22S2*d2S2 + d22S3*d2S3);

  // computing iIS &* transpose(iIS)

  gg11 = g11*g11 + g12*g12;
  gg12 = g11*g12 + g12*g22;
  gg22 = g12*g12 + g22*g22;

  // computing DomegaS

  DomegaS=sqrt(det_IS);

  // compare stuff

  set_Hess_S_of_h(u,v,
		  d1h[0],d2h[0],
		  d11h[0],d12h[0],d22h[0]);
  /*
  printf("inv_IS  %e %e %e %e\n",inv_IS[0][0]-g11,
          	                 inv_IS[0][1]-g12,
          	                 inv_IS[1][0]-g12,
  	                         inv_IS[1][1]-g22);

  printf("Gamma1  %e %e %e %e\n",Gamma[0][0][0]-Gamma1_11,
          	                 Gamma[0][1][0]-Gamma1_12,
          	                 Gamma[1][0][0]-Gamma1_12,
  	                         Gamma[1][1][0]-Gamma1_22);

  printf("Gamma2  %e %e %e %e\n",Gamma[0][0][1]-Gamma2_11,
          	                 Gamma[0][1][1]-Gamma2_12,
          	                 Gamma[1][0][1]-Gamma2_12,
  	                         Gamma[1][1][1]-Gamma2_22);

  printf("Domega  %e\n",DomegaS-Domega);
  */

  if(1)
    {
      for(i=0;i<16;i++)
	for(j=0;j<16;j++)
	  {
	    double t0;

	    t0 = 
	      ((d11h[i]-d1h[i]*Gamma1_11-d2h[i]*Gamma2_11)*gg11+
	       (d12h[i]-d1h[i]*Gamma1_12-d2h[i]*Gamma2_12)*gg12)*
	       (d11h[j]-d1h[j]*Gamma1_11-d2h[j]*Gamma2_11)+

	      ((d11h[i]-d1h[i]*Gamma1_11-d2h[i]*Gamma2_11)*gg12+
	       (d12h[i]-d1h[i]*Gamma1_12-d2h[i]*Gamma2_12)*gg22)*
	       (d12h[j]-d1h[j]*Gamma1_12-d2h[j]*Gamma2_12)+

	      ((d12h[i]-d1h[i]*Gamma1_12-d2h[i]*Gamma2_12)*gg11+
	       (d22h[i]-d1h[i]*Gamma1_22-d2h[i]*Gamma2_22)*gg12)*
	       (d12h[j]-d1h[j]*Gamma1_12-d2h[j]*Gamma2_12)+

	      ((d12h[i]-d1h[i]*Gamma1_12-d2h[i]*Gamma2_12)*gg12+
	       (d22h[i]-d1h[i]*Gamma1_22-d2h[i]*Gamma2_22)*gg22)*
	       (d22h[j]-d1h[j]*Gamma1_22-d2h[j]*Gamma2_22);

	    K[i*16+j]+=weight*t0*DomegaS;
	    //K[j*16+i]+=weight*t0*DomegaS;
	  }
    }
  else
    {
      for(i=1;i<16;i++)
	for(j=i;j<16;j++)
	  {
	    double t0;//,t3;//,t7,t8,t19,t24;
	  }
    }
}

//////////////////////////////////////////////////////////////////////////////

void BicubicPatch::add_sample_to_Km(double u, double v, double weight, double *K)
{
  double tmp;
  tmp=u+v+weight+(int)K;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void BicubicPatch::Kp(Patch *S, double *K)
{
  int    i, j, k;
  double t;
  double *tmp;
  tmp=new double[256];

  assert(S==NULL);

  for(i=0;i<256;i++) K[i]=0.0;

  //add_sample_to_Kp(0.5, 0.5, 1.0, K);

  //add_sample_to_Kp(0.0, 0.0, 0.25, K);
  //add_sample_to_Kp(0.0, 1.0, 0.25, K);
  //add_sample_to_Kp(1.0, 0.0, 0.25, K);
  //add_sample_to_Kp(1.0, 1.0, 0.25, K);

  add_sample_to_Kp(    0.0,     0.0, 1.0/64.0, K);
  add_sample_to_Kp(    0.0, 1.0/3.0, 3.0/64.0, K);
  add_sample_to_Kp(    0.0, 2.0/3.0, 3.0/64.0, K);
  add_sample_to_Kp(    0.0,     1.0, 1.0/64.0, K);

  add_sample_to_Kp(1.0/3.0,     0.0, 3.0/64.0, K);
  add_sample_to_Kp(1.0/3.0, 1.0/3.0, 9.0/64.0, K);
  add_sample_to_Kp(1.0/3.0, 2.0/3.0, 9.0/64.0, K);
  add_sample_to_Kp(1.0/3.0,     1.0, 3.0/64.0, K);

  add_sample_to_Kp(2.0/3.0,     0.0, 3.0/64.0, K);
  add_sample_to_Kp(2.0/3.0, 1.0/3.0, 9.0/64.0, K);
  add_sample_to_Kp(2.0/3.0, 2.0/3.0, 9.0/64.0, K);
  add_sample_to_Kp(2.0/3.0,     1.0, 3.0/64.0, K);

  add_sample_to_Kp(    1.0,     0.0, 1.0/64.0, K);
  add_sample_to_Kp(    1.0, 1.0/3.0, 3.0/64.0, K);
  add_sample_to_Kp(    1.0, 2.0/3.0, 3.0/64.0, K);
  add_sample_to_Kp(    1.0,     1.0, 1.0/64.0, K);

  for(i=0;i<16;++i)
    {
      for(k=0;k<16;++k) 
	{
	  t=0.0;
	  for(j=0;j<16;++j )
	    {
	      t+=K[i*16+j]*iX_patrick[j*16+k];
	    }
	  tmp[i*16+k]=t;
	}
    }
  for(i=0;i<16;++i)
    {
      for(k=0;k<16;++k) 
	{
	  t=0.0;
	  for(j=0;j<16;++j )
	    {
	      t+=iX_patrick[j*16+i]*tmp[j*16+k];
	    } //             ^^^^^ j and i transposed
	  K[i*16+k]=t;
	}
    }
}

//////////////////////////////////////////////////////////////////////////////

void BicubicPatch::Km(Patch *S, double *K)
{
  assert(S==NULL && K==NULL);

}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void BicubicPatch::load_matrix(char *fname, double *p, int n, int m)
{
  int  i,j;
  FILE *f;

  //printf("%s  %d,%d\n",fname,n,m);

  f=fopen(fname,"r");
  if(!f)
    {
      cerr << "Error: File '"<<fname<<"' doesn't exist! (load_matrix())" <<endl;
      exit(1);
    }

  for(i=0;i<n;i++)
    {
      for(j=0;j<m;j++)
	{
	  fscanf(f,"%lf",&(p[i*m+j]));
	}
      fscanf(f,"\n");
    }
  fclose(f);
}

//////////////////////////////////////////////////////////////////////////////

void BicubicPatch::set_matrices(void)
{
  load_matrix("data/regular.floatingpoint/regular_eX_patrick.mtx",eX_patrick,16,16);
  load_matrix("data/regular.floatingpoint/regular_iX_patrick.mtx",iX_patrick,16,16);
  load_matrix("data/regular.floatingpoint/regular_eX_tensor.mtx", eX_tensor, 16,16);
  load_matrix("data/regular.floatingpoint/regular_iX_tensor.mtx", iX_tensor, 16,16);
  load_matrix("data/regular.floatingpoint/regular_Lambda.mtx",    Lambda,    16,16);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

//double BicubicPatch::grad_KpEnergy(int coord, int i)
//{
//
//}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
