//////////////////////////////////////////////////////////////////////////////
//
// Integrate.C, 07/15/01, Ilja Friedel
//
//
//////////////////////////////////////////////////////////////////////////////
 
#include "Patch.h"

#define INTEGRATION_ACCURACY 1.0e-2

#ifdef OLD_INTEGRATE
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

double Patch::GaussInt(F2D f, double a, double b, double c, double d)
{
  const double GAUSS_X[] = {0., -.9061798459386638, .9061798459386638,
			    -.5384693101056829, .5384693101056829};

  const double GAUSS_W[] = {.5688888888888889, .2369268850561895, 
			    .2369268850561895,
			    .4786286704993660, .4786286704993660};

  int i,j;
  double s,bma2,apb2,dmc2,cpd2;

  s=0;
  bma2=(b-a)*0.5;
  apb2=(a+b)*0.5;
  dmc2=(d-c)*0.5;
  cpd2=(c+d)*0.5;
  
  for(i=0;i<5;i++)
    for(j=0;j<5;j++)
      {
	s=s+GAUSS_W[i]*GAUSS_W[j]*
	  (*this.*f)(bma2*GAUSS_X[i]+apb2,dmc2*GAUSS_X[j]+cpd2);
      }
  
  return(bma2*dmc2*s);
}

//////////////////////////////////////////////////////////////////////////////

double Patch::NumInt(F2D f,
		     double a, double b, double c, double d, 
		     int N)
{
  int i,j;
  double sum,dx,dy;
  sum=0.0;
  
  dx=(b-a)/N;
  dy=(d-c)/N;
 
  for(i=0;i<N;i++)
    {
      for(j=0;j<N;j++)
	{
	  sum=sum+GaussInt(f,a+i*dx,a+(i+1)*dx,c+j*dy,c+(j+1)*dy);
	}
    }

   return(sum);
}

//////////////////////////////////////////////////////////////////////////////

double Patch::Integrate(F2D f, double a, double b,
			double c, double d)
{
  double I,I_old;
  int N;

  I_old = NumInt(f, a, b, c, d, 1);
  I     = NumInt(f, a, b, c, d, 2);
  
  N=2;
  while(fabs(I-I_old)/(1.0+0.5*(fabs(I)+fabs(I_old)))>1e-2)
    {
      N=2*N;
      if(N>=16) printf("Integrating adaptivly! Level %d. (I=%1.17e, I_old=%1.17e)\n",N,I,I_old);
      
      I_old = I;
      I     = NumInt(f, a, b, c, d, N);
    }
  
  return(I);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
#else

double Patch::GaussInt(F2D f, double a, double b, double c, double d, int order)
{
  const double GAUSS_X[5][5] = 
    {
      {0.0,                                                                           0.0, 0.0, 0.0, 0.0}, 
      {.57735026918962576449, -.57735026918962576449,                                      0.0, 0.0, 0.0}, 
      {0., .77459666924148337704, -.77459666924148337704,                                       0.0, 0.0}, 
      {-.86113631159405257524, .86113631159405257524, -.33998104358485626481, .33998104358485626481, 0.0},
      {0., -.9061798459386638, .9061798459386638, -.5384693101056829, .5384693101056829}
    };

  const double GAUSS_W[5][5] = 
    {
      {2.0,                                                                         0.0, 0.0, 0.0, 0.0},
      {1.0000000000000000000, 1.0000000000000000000,                                     0.0, 0.0, 0.0},
      {.88888888888888888889, .55555555555555555552, .55555555555555555552,                   0.0, 0.0},
      {.34785484513745385737, .34785484513745385737, .65214515486254614266, .65214515486254614266, 0.0},
      {.5688888888888889, .2369268850561895, .2369268850561895, .4786286704993660, .4786286704993660}
    };

  int i,j;
  double s,bma2,apb2,dmc2,cpd2;

  assert(order>0 && order <6);

  s=0.0;
  bma2=(b-a)*0.5;
  apb2=(a+b)*0.5;
  dmc2=(d-c)*0.5;
  cpd2=(c+d)*0.5;

  if(order==0)  // Monte Carlo Integration, not working
    {
      for(i=0;i<16;i++)
	{
	  cerr << "Error: MC integration not implemented." << endl;
	  exit(1);
	  //	    s=s+4.0*
	  //(*this.*f)(bma2*2.0*(drand48()-0.5)+apb2,
	  //		 dmc2*2.0*(drand48()-0.5)+cpd2);
	}
      s=s/16.0;
    }
  else
    {
      for(i=0;i<order;i++)
	for(j=0;j<order;j++)
	  {
	    s=s+
	      GAUSS_W[order-1][i]*
	      GAUSS_W[order-1][j]*
	      (*this.*f)(bma2*GAUSS_X[order-1][i]+apb2,
			 dmc2*GAUSS_X[order-1][j]+cpd2);
	  }
    }
  return(bma2*dmc2*s);
}

//////////////////////////////////////////////////////////////////////////////

double Patch::NumInt(F2D f,
		     double a, double b, double c, double d, 
		     double I_old,
		     double accuracy, int level)
{
  double I1, I2, I3, I4;
  double I;
  int    order;

  //if(accuracy<1e-10) accuracy=1e-10;  // ***********************************************

  order=2;
  //if(level>5 ) order=4;
  //if(level>10) order=3;
  //if(level>15) order=2;
  if(level>20) order=1;
  if(level>30)
    {
      /*
      cerr << "Warning: Can't reach integration accuracy! (level="<<level<<")"<<endl;
      cerr << "         I_old "<<I_old<<"   Accuracy "<<accuracy<<endl;
      cerr << "         Possibly discontinuity in ["<<a<<","<<b<<"] x ["<<c<<","<<d<<"]"<<endl;
      */
      return(I_old);
    }

  if(order<4)
    {
      int    N;
      int    i, j;
      double dx, dy;

      I1 = 0.0;
      I2 = 0.0;
      I3 = 0.0;
      I4 = 0.0;
      
      N=2;
      if(order==1) N=5;
      if(order==2) N=3;

      dx=0.5*(b-a)/N;
      dy=0.5*(d-c)/N;

      for(i=0;i<N;i++)
	for(j=0;j<N;j++)
	  {
	    I1   += GaussInt(f, a +              i   *dx, 
			        a +             (i+1)*dx,
			        c +              j   *dy, 
			        c +             (j+1)*dy, order);

	    I2   += GaussInt(f, a +              i   *dx, 
			        a +             (i+1)*dx,
			        c + (d-c)*0.5 +  j   *dy, 
			        c + (d-c)*0.5 + (j+1)*dy, order);

	    I3   += GaussInt(f, a + (b-a)*0.5 +  i   *dx,
			        a + (b-a)*0.5 + (i+1)*dx,
			        c +              j   *dy, 
			        c +             (j+1)*dy, order);

	    I4   += GaussInt(f, a + (b-a)*0.5 +  i   *dx,
			        a + (b-a)*0.5 + (i+1)*dx,
			        c + (d-c)*0.5 +  j   *dy, 
			        c + (d-c)*0.5 + (j+1)*dy, order);
	  }
    }
  else
    {
      I1    = GaussInt(f, a,             a + (b-a)*0.5, c,             c + (d-c)*0.5, order);
      I2    = GaussInt(f, a,             a + (b-a)*0.5, c + (d-c)*0.5, d,             order);
      I3    = GaussInt(f, a + (b-a)*0.5, b,             c,             c + (d-c)*0.5, order);
      I4    = GaussInt(f, a + (b-a)*0.5, b,             c + (d-c)*0.5, d,             order);
    }

  I     = I1 + I2 + I3 + I4;

  if(fabs(I-I_old)/(1.0+0.5*(fabs(I)+fabs(I_old)))>accuracy)
    {
      I=NumInt(f, a, a + (b-a)*0.5,  c, c + (d-c)*0.5,  I1, 0.5*accuracy, level+1)+ // technically 0.25*accuracy  *******
	NumInt(f, a, a + (b-a)*0.5,  c + (d-c)*0.5, d,  I2, 0.5*accuracy, level+1)+
	NumInt(f, a + (b-a)*0.5, b,  c, c + (d-c)*0.5,  I3, 0.5*accuracy, level+1)+
	NumInt(f, a + (b-a)*0.5, b,  c + (d-c)*0.5, d,  I4, 0.5*accuracy, level+1);
    }

  return(I);  
}

//////////////////////////////////////////////////////////////////////////////

double Patch::Integrate(F2D f, 
			double a, double b,
			double c, double d)
{
  double I,I_old;
  double I1, I2, I3, I4;
  int    order;

  order = 3;

  I_old = GaussInt(f, a, b, c, d, order);

  I1    = GaussInt(f, a,             a + (b-a)*0.5, c,             c + (d-c)*0.5, order);
  I2    = GaussInt(f, a,             a + (b-a)*0.5, c + (d-c)*0.5, d,             order);
  I3    = GaussInt(f, a + (b-a)*0.5, b,             c,             c + (d-c)*0.5, order);
  I4    = GaussInt(f, a + (b-a)*0.5, b,             c + (d-c)*0.5, d,             order);

  I     = I1 + I2 + I3 + I4;
  
  if(fabs(I-I_old)/(1.0e-9+0.5*(fabs(I)+fabs(I_old)))>INTEGRATION_ACCURACY)
    {
      I=NumInt(f, a, a + (b-a)*0.5, c, c + (d-c)*0.5, I1, 0.25*INTEGRATION_ACCURACY, 1)+
	NumInt(f, a, a + (b-a)*0.5, c + (d-c)*0.5, d, I2, 0.25*INTEGRATION_ACCURACY, 1)+
	NumInt(f, a + (b-a)*0.5, b, c, c + (d-c)*0.5, I3, 0.25*INTEGRATION_ACCURACY, 1)+
	NumInt(f, a + (b-a)*0.5, b, c + (d-c)*0.5, d, I4, 0.25*INTEGRATION_ACCURACY, 1);
    }

  return(I);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#endif
