//////////////////////////////////////////////////////////////////////////////
//
// EigenStructure.C, 07/18/01, Ilja Friedel
//
//
//////////////////////////////////////////////////////////////////////////////

#include "EigenStructure.h"
#include "ExtraordinaryPatch.h"

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

BicubicPatch **EigenBasis[MAX_VALENCE+1][3];
int EigenStructure::initialized=0;

double *iX [MAX_VALENCE+1]; // inverse eigenvectors
double *eX [MAX_VALENCE+1]; // eigenvectors

double *Lambda[MAX_VALENCE+1]; 
double *Omega1[MAX_VALENCE+1];
double *Omega2[MAX_VALENCE+1];
double *Omega3[MAX_VALENCE+1];
double *Omega4[MAX_VALENCE+1];

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

EigenStructure::EigenStructure(Patch *aEP)
{
  EP = aEP;
  if(EP==NULL)
    {
      cerr << "Error: Patch for EigenStructure is NULL!"<<endl;
      exit(1);
    }
  //EP->print();
  N  = ((ExtraordinaryPatch*)EP)->Valence;
  if(N<3 || N>60)
    {
      printf("Error: Valence N=%d out of range.\n",N);
      exit(1);
    }

  M=((ExtraordinaryPatch*)EP)->M;

  if(M!=2*N+8) 
    {
      cerr << "Error: M!=2*Valence+8!"<<endl;
      exit(1);
    }

  if(!initialized)
    {
      int i;
      cout << "Initializing iX, eX, Lambda and Omegas." << endl;

      for(i=3;i<MAX_VALENCE+1;i++)
	{
	  iX[N]=NULL;
	  eX[N]=NULL;

	  Lambda[N]=NULL;
	  Omega1[N]=NULL;
	  Omega2[N]=NULL;
	  Omega3[N]=NULL;
	  Omega4[N]=NULL;
	}
      initialized=1;
    }

  if(!iX[N])
    {
      iX[N]     = new double[M*M];
      eX[N]     = new double[M*M];

      Lambda[N] = new double[M*M];
      Omega1[N] = new double[16*M];
      Omega2[N] = new double[16*M];
      Omega3[N] = new double[16*M];
      Omega4[N] = new double[M*M];

      if(!(iX[N]&&eX[N]&&Lambda[N]&&Omega1[N]&&Omega2[N]&&Omega3[N]&&Omega4[N]))
	{
	  printf("Error: Out of memory!\n");
	  exit(1);
	}

      load_all_general_matrices();
      create_EigenFunctions();
    }
  project_controlpoints_to_eigenbasis();
}

//////////////////////////////////////////////////////////////////////////////
// ok
void EigenStructure::project_controlpoints_to_eigenbasis(void)
{
  int i,j;
  
  eigen_controlpoints[0]=new double[M];
  eigen_controlpoints[1]=new double[M];
  eigen_controlpoints[2]=new double[M];

  assert(eigen_controlpoints[0]);
  assert(eigen_controlpoints[1]);
  assert(eigen_controlpoints[2]);

  //cout << endl << "Valence="<<(M-8)/2<<endl;
  for(i=0;i<M;i++)
    {
      eigen_controlpoints[0][i]=0.0;
      eigen_controlpoints[1][i]=0.0;
      eigen_controlpoints[2][i]=0.0;

      for(j=0;j<M;j++)
	{
	  //cout << iX[N][i*M+j] << " ";

	  eigen_controlpoints[0][i] +=iX[N][i*M+j] * ((ExtraordinaryPatch*)EP)->control_points[0][j]; 
	  eigen_controlpoints[1][i] +=iX[N][i*M+j] * ((ExtraordinaryPatch*)EP)->control_points[1][j]; 
	  eigen_controlpoints[2][i] +=iX[N][i*M+j] * ((ExtraordinaryPatch*)EP)->control_points[2][j]; 
	}
      //cout <<endl;
      //cout << "Cp["<<i<<"] = ["
      //<<clip(eigen_controlpoints[0][i])<<", "
      //<<clip(eigen_controlpoints[1][i])<<", "
      //<<clip(eigen_controlpoints[2][i])<<"]" << endl;
    }
  //cout <<endl;
}

//////////////////////////////////////////////////////////////////////////////
// ok
void EigenStructure::create_EigenFunctions(void)
{
  int    i,j,k;
  double EigenFunctionControlPoints[16]; 

  assert(Omega1[N] && Omega2[N] && Omega3[N]);
  
  EigenBasis[N][0]=new BicubicPatch*[M];
  EigenBasis[N][1]=new BicubicPatch*[M];
  EigenBasis[N][2]=new BicubicPatch*[M];

  assert(EigenBasis[N][0]);
  assert(EigenBasis[N][1]);
  assert(EigenBasis[N][2]);

  for(i=0;i<M;i++)
    {
      for(j=0;j<16;j++)
	{
	  EigenFunctionControlPoints[j]=0.0;
	  for(k=0;k<M;k++) EigenFunctionControlPoints[j] += Omega1[N][j*M+k] * eX[N][k*M+i];
	}
      EigenBasis[N][0][i]=new BicubicPatch(EigenFunctionControlPoints);

      for(j=0;j<16;j++)
	{
	  EigenFunctionControlPoints[j]=0.0;
	  for(k=0;k<M;k++) EigenFunctionControlPoints[j] += Omega2[N][j*M+k] * eX[N][k*M+i];
	}
      EigenBasis[N][1][i]=new BicubicPatch(EigenFunctionControlPoints);

      for(j=0;j<16;j++)
	{
	  EigenFunctionControlPoints[j]=0.0;
	  for(k=0;k<M;k++) EigenFunctionControlPoints[j] += Omega3[N][j*M+k] * eX[N][k*M+i];
	}
      EigenBasis[N][2][i]=new BicubicPatch(EigenFunctionControlPoints);
    }
}

//////////////////////////////////////////////////////////////////////////////

EigenStructure::~EigenStructure()
{
  if(eigen_controlpoints[0]) delete[] eigen_controlpoints[0];
  if(eigen_controlpoints[1]) delete[] eigen_controlpoints[1];
  if(eigen_controlpoints[2]) delete[] eigen_controlpoints[2];

  eigen_controlpoints[0]=NULL;
  eigen_controlpoints[1]=NULL;
  eigen_controlpoints[2]=NULL;

  // don't delete EigenFunctions, they are global
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

double EigenStructure::evaluate_phi(double u, double v, int mode)
{
  const  double ilog2=1.0/log(2.0);
  //const  double EPSILON=pow(2.0,-300);
  double tmp;
  double n1,n2;
  int    n,k;
  double pow2;

  assert(u>=0.0 && u<=1.0); // ****** see derivatives
  assert(v>=0.0 && v<=1.0);
  assert(mode>=0&& mode<M);

  if(u<EPSILON1) u=EPSILON1;
  if(v<EPSILON1) v=EPSILON1;

  // we allow using the interval [0,1] instead of [0,1)
  if(u==1.0) n1=1.0;
  else       n1=-log(u)*ilog2;
  if(v==1.0) n2=1.0;
  else       n2=-log(v)*ilog2;

  if(n1<n2) n=(int)(ceil(n1)); 
  else      n=(int)(ceil(n2));

  pow2=pow(2.0,n-1);
  u = u*pow2;
  v = v*pow2;

  // we compute v=1.0-... and u=1.0-... because the parametrizations
  // for bicubic patches and extraordinary patches were chosen in oposite
  // directions, e.g. (X denoting origin)
  //
  // bicubic:         0  1  2  3 
  //                  4  X  6  7
  //                  8  9 10 11
  //                 12 13 14 15
  //
  // extraordinary:   0  1  3  4
  //                  5  6  7  8 
  //                  9 10  X 11
  //                 12 13 14 15 
  //
  // See also Maple code:
  //
  // if   l=1 then Omega:=matrix(16,2*N+8,curry(Omega1,N)); 
  // U:=matrix(1,4,[(2-2*u)^3,(2-2*u)^2,(2-2*u),1]); 
  // V:=matrix(1,4,[(2-2*v)^3,(2-2*v)^2,(2-2*v),1]);   
  //
  // elif l=2 then Omega:=matrix(16,2*N+8,curry(Omega2,N)); 
  // U:=matrix(1,4,[(2-2*u)^3,(2-2*u)^2,(2-2*u),1]); 
  // V:=matrix(1,4,[(1-2*v)^3,(1-2*v)^2,(1-2*v),1]);
  //
  // elif l=3 then Omega:=matrix(16,2*N+8,curry(Omega3,N)); 
  // U:=matrix(1,4,[(1-2*u)^3,(1-2*u)^2,(1-2*u),1]); 
  // V:=matrix(1,4,[(2-2*v)^3,(2-2*v)^2,(2-2*v),1]); 


  if(u<0.5)
    {
      assert(v>=0.5 && v<=1.0);

      k=2;
      v=1.0-(2.0 * v - 1.0);
      u=1.0-(2.0 * u);
    }
  else if(v<0.5)
    {
      assert(u>=0.5 && u<=1.0);

      k=1;
      v=1.0-(2.0 * v);
      u=1.0-(2.0 * u - 1.0);
    }
  else 
    {
      assert(v>=0.5 && v<=1.0);
      assert(u>=0.5 && u<=1.0);

      k=0;
      u=1.0-(2.0 * u - 1.0);
      v=1.0-(2.0 * v - 1.0);
    }

  //  if(!s) cout <<"New u,v is "<< u<<", "<<v<< "   (n1, n2, n)=("<<(int)(ceil(n1))<<", "
  //	      <<(int)(ceil(n2))<<", "<<n<<")   k="<<k<<endl;

  tmp=pow(Lambda[N][mode*M+mode],n-1) * EigenBasis[N][k][mode]->evaluate_S(u,v,0); 

  return(tmp);
}

//////////////////////////////////////////////////////////////////////////////

double EigenStructure::evaluate_D1phi(double u, double v, int mode)
{
  const  double ilog2=1.0/log(2.0);
  //  const  double EPSILON=pow(2.0,-300);
  double tmp;
  double n1,n2;
  int    n,k;
  double pow2;

  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);  
  assert(mode>=0&& mode<M);

  if(u<EPSILON1) u=EPSILON1;
  if(v<EPSILON1) v=EPSILON1;

  // we allow using the interval [0,1] instead of [0,1)
  if(u==1.0) n1=1.0;
  else       n1=-log(u)*ilog2;
  if(v==1.0) n2=1.0;
  else       n2=-log(v)*ilog2;

  if(n1<n2) n=(int)(ceil(n1)); 
  else      n=(int)(ceil(n2));

  pow2=pow(2.0,n-1);
  u = u*pow2;
  v = v*pow2;

  if(u<0.5)
    {
      assert(v>=0.5 && v<=1.0);

      k=2;
      v=1.0-(2.0 * v - 1.0);
      u=1.0-(2.0 * u);
    }
  else if(v<0.5)
    {
      assert(u>=0.5 && u<=1.0);

      k=1;
      v=1.0-(2.0 * v);
      u=1.0-(2.0 * u - 1.0);
    }
  else 
    {
      assert(v>=0.5 && v<=1.0);
      assert(u>=0.5 && u<=1.0);

      k=0;
      u=1.0-(2.0 * u - 1.0);
      v=1.0-(2.0 * v - 1.0);
    }

  tmp=pow(Lambda[N][mode*M+mode],n-1) * EigenBasis[N][k][mode]->evaluate_D1S(u,v,0); 

  return(pow(2.0,n)*tmp);
}

//////////////////////////////////////////////////////////////////////////////

double EigenStructure::evaluate_D2phi(double u, double v, int mode)
{
  const  double ilog2=1.0/log(2.0);
  //  const  double EPSILON=pow(2.0,-300);
  double tmp;
  double n1,n2;
  int    n,k;
  double pow2;

  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);
  assert(mode>=0&& mode<M);

  if(u<EPSILON1) u=EPSILON1;
  if(v<EPSILON1) v=EPSILON1;

  // we allow using the interval [0,1] instead of [0,1)
  if(u==1.0) n1=1.0;
  else       n1=-log(u)*ilog2;
  if(v==1.0) n2=1.0;
  else       n2=-log(v)*ilog2;

  if(n1<n2) n=(int)(ceil(n1)); 
  else      n=(int)(ceil(n2));

  pow2=pow(2.0,n-1);
  u = u*pow2;
  v = v*pow2;

  if(u<0.5)
    {
      assert(v>=0.5 && v<=1.0);

      k=2;
      v=1.0-(2.0 * v - 1.0);
      u=1.0-(2.0 * u);
    }
  else if(v<0.5)
    {
      assert(u>=0.5 && u<=1.0);

      k=1;
      v=1.0-(2.0 * v);
      u=1.0-(2.0 * u - 1.0);
    }
  else 
    {
      assert(v>=0.5 && v<=1.0);
      assert(u>=0.5 && u<=1.0);

      k=0;
      u=1.0-(2.0 * u - 1.0);
      v=1.0-(2.0 * v - 1.0);
    }

  tmp=pow(Lambda[N][mode*M+mode],n-1) * EigenBasis[N][k][mode]->evaluate_D2S(u,v,0); 

  return(pow(2.0,n)*tmp);
}

//////////////////////////////////////////////////////////////////////////////

double EigenStructure::evaluate_D1D1phi(double u, double v, int mode)
{
  const  double ilog2=1.0/log(2.0);
  //  const  double EPSILON=pow(2.0,-300);
  double tmp;
  double n1,n2;
  int    n,k;
  double pow2;

  assert(u> 0.0 || v> 0.0);
  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);
  assert(mode>=0&& mode<M);

  if(u<EPSILON1) u=EPSILON1;
  if(v<EPSILON1) v=EPSILON1;

  // we allow using the interval [0,1] instead of [0,1)
  if(u==1.0) n1=1.0;
  else       n1=-log(u)*ilog2;
  if(v==1.0) n2=1.0;
  else       n2=-log(v)*ilog2;

  if(n1<n2) n=(int)(ceil(n1)); 
  else      n=(int)(ceil(n2));

  pow2=pow(2.0,n-1);
  u = u*pow2;
  v = v*pow2;

  if(u<0.5)
    {
      assert(v>=0.5 && v<=1.0);

      k=2;
      v=1.0-(2.0 * v - 1.0);
      u=1.0-(2.0 * u);
    }
  else if(v<0.5)
    {
      assert(u>=0.5 && u<=1.0);

      k=1;
      v=1.0-(2.0 * v);
      u=1.0-(2.0 * u - 1.0);
    }
  else 
    {
      assert(v>=0.5 && v<=1.0);
      assert(u>=0.5 && u<=1.0);

      k=0;
      u=1.0-(2.0 * u - 1.0);
      v=1.0-(2.0 * v - 1.0);
    }

  tmp=pow(Lambda[N][mode*M+mode],n-1) * EigenBasis[N][k][mode]->evaluate_D1D1S(u,v,0); 

  return(pow(2.0,2.0*n)*tmp);
}

//////////////////////////////////////////////////////////////////////////////

double EigenStructure::evaluate_D1D2phi(double u, double v, int mode)
{
  const  double ilog2=1.0/log(2.0);
  //  const  double EPSILON=pow(2.0,-300);
  double tmp;
  double n1,n2;
  int    n,k;
  double pow2;

  assert(u> 0.0 || v> 0.0);
  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);  
  assert(mode>=0&& mode<M);

  if(u<EPSILON1) u=EPSILON1;
  if(v<EPSILON1) v=EPSILON1;

  // we allow using the interval [0,1] instead of [0,1)
  if(u==1.0) n1=1.0;
  else       n1=-log(u)*ilog2;
  if(v==1.0) n2=1.0;
  else       n2=-log(v)*ilog2;

  if(n1<n2) n=(int)(ceil(n1)); 
  else      n=(int)(ceil(n2));

  pow2=pow(2.0,n-1);
  u = u*pow2;
  v = v*pow2;

  if(u<0.5)
    {
      assert(v>=0.5 && v<=1.0);

      k=2;
      v=1.0-(2.0 * v - 1.0);
      u=1.0-(2.0 * u);
    }
  else if(v<0.5)
    {
      assert(u>=0.5 && u<=1.0);

      k=1;
      v=1.0-(2.0 * v);
      u=1.0-(2.0 * u - 1.0);
    }
  else 
    {
      assert(v>=0.5 && v<=1.0);
      assert(u>=0.5 && u<=1.0);

      k=0;
      u=1.0-(2.0 * u - 1.0);
      v=1.0-(2.0 * v - 1.0);
    }

  tmp=pow(Lambda[N][mode*M+mode],n-1) * EigenBasis[N][k][mode]->evaluate_D1D2S(u,v,0); 

  return(pow(2.0,2.0*n)*tmp);
}

//////////////////////////////////////////////////////////////////////////////

double EigenStructure::evaluate_D2D1phi(double u, double v, int mode)
{
  return(evaluate_D1D2phi(u,v,mode));
}

//////////////////////////////////////////////////////////////////////////////

double EigenStructure::evaluate_D2D2phi(double u, double v, int mode)
{
  const  double ilog2=1.0/log(2.0);
  //  const  double EPSILON=pow(2.0,-300);
  double tmp;
  double n1,n2;
  int    n,k;
  double pow2;

  assert(u> 0.0 || v> 0.0);
  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);
  assert(mode>=0&& mode<M);

  if(u<EPSILON1) u=EPSILON1;
  if(v<EPSILON1) v=EPSILON1;

  // we allow using the interval [0,1] instead of [0,1)
  if(u==1.0) n1=1.0;
  else       n1=-log(u)*ilog2;
  if(v==1.0) n2=1.0;
  else       n2=-log(v)*ilog2;

  if(n1<n2) n=(int)(ceil(n1)); 
  else      n=(int)(ceil(n2));

  pow2=pow(2.0,n-1);
  u = u*pow2;
  v = v*pow2;

  if(u<0.5)
    {
      assert(v>=0.5 && v<=1.0);

      k=2;
      v=1.0-(2.0 * v - 1.0);
      u=1.0-(2.0 * u);
    }
  else if(v<0.5)
    {
      assert(u>=0.5 && u<=1.0);

      k=1;
      v=1.0-(2.0 * v);
      u=1.0-(2.0 * u - 1.0);
    }
  else 
    {
      assert(v>=0.5 && v<=1.0);
      assert(u>=0.5 && u<=1.0);

      k=0;
      u=1.0-(2.0 * u - 1.0);
      v=1.0-(2.0 * v - 1.0);
    }

  tmp=pow(Lambda[N][mode*M+mode],n-1) * EigenBasis[N][k][mode]->evaluate_D2D2S(u,v,0); 

  return(pow(2.0,2.0*n)*tmp);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void EigenStructure::load_matrix(char *fname, double *p, int n, int m)
{
  int  i,j;
  FILE *f;

  printf("%s  %d,%d\n",fname,n,m);

  f=fopen(fname,"r");
  if(!f)
    {
      cerr << "Error: File '"<<fname<<"' doesn't exist! (load_matrix())" <<endl;
      exit(1);
    }

  for(i=0;i<n;i++)
    {
      for(j=0;j<m;j++)
	{
	  char c;
	  double q;

	  fscanf(f,"%lf",&(p[i*m+j]));
	  // reading quotients
	  c=getc(f);
	  if(c=='/')
	    {
	      fscanf(f,"%lf",&q);
	      assert(q!=0.0);
	      p[i*m+j]/=q;
	    }
	  else ungetc(c,f);

	}
      fscanf(f,"\n");
    }
  fclose(f);
}

//////////////////////////////////////////////////////////////////////////////

void EigenStructure::save_matrix(char *fname, double *p, int n, int m)
{
  int i,j;
  FILE *f;

  f=fopen(fname,"w");
  for(i=0;i<n;i++)
    {
      for(j=0;j<m;j++)
	{
	  fprintf(f,"%1.10e\t",p[i*m+j]);
	}
      fprintf(f,"\n");
    }
  fclose(f);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void EigenStructure::load_all_general_matrices(void)
{
  char fname[1024];
  const char path[128]="./raw_matrices.sorted";

  printf("Loading general matrices ... ");
  fflush(0);

  if(N<10)
    {	
      sprintf(fname,"%s/0%d/N%d_iX.mtx",path,N,N);
      load_matrix(fname,iX[N],M,M);
      sprintf(fname,"%s/0%d/N%d_eX.mtx",path,N,N);
      load_matrix(fname,eX[N],M,M);

      sprintf(fname,"%s/0%d/N%d_Lambda.mtx",path,N,N);
      load_matrix(fname,Lambda[N],M,M);

      sprintf(fname,"%s/0%d/N%d_Omega1.mtx",path,N,N);
      load_matrix(fname,Omega1[N],16,M);
      sprintf(fname,"%s/0%d/N%d_Omega2.mtx",path,N,N);
      load_matrix(fname,Omega2[N],16,M);
      sprintf(fname,"%s/0%d/N%d_Omega3.mtx",path,N,N);
      load_matrix(fname,Omega3[N],16,M);
      sprintf(fname,"%s/0%d/N%d_Omega4.mtx",path,N,N);
      load_matrix(fname,Omega4[N],M,M);
    }
  else
    {
      sprintf(fname,"%s/%d/N%d_iX.mtx",path,N,N);
      load_matrix(fname,iX[N],M,M);
      sprintf(fname,"%s/%d/N%d_eX.mtx",path,N,N);
      load_matrix(fname,eX[N],M,M);

      sprintf(fname,"%s/%d/N%d_Lambda.mtx",path,N,N);
      load_matrix(fname,Lambda[N],M,M);

      sprintf(fname,"%s/%d/N%d_Omega1.mtx",path,N,N);
      load_matrix(fname,Omega1[N],16,M);
      sprintf(fname,"%s/%d/N%d_Omega2.mtx",path,N,N);
      load_matrix(fname,Omega2[N],16,M);
      sprintf(fname,"%s/%d/N%d_Omega3.mtx",path,N,N);
      load_matrix(fname,Omega3[N],16,M);
      sprintf(fname,"%s/%d/N%d_Omega4.mtx",path,N,N);
      load_matrix(fname,Omega3[N],M,M);
    }

  printf("done!\n");
  fflush(0);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void EigenStructure::matrix_mult( double *A, double *B, double *C, int I, int J, int K)
{
  int    i, j, k;
  double t;

  for(i=0;i<I;++i)
    {
      for(k=0;k<K;++k) 
	{
	  t=0.0;
	  for(j=0;j<J;++j )
	    {
	      t+=A[i*J+j]*B[j*K+k];
	    }
	  C[i*K+k]=t;
	}
    }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

