//////////////////////////////////////////////////////////////////////////////
//
// Patch.h, 07/24/01, Ilja Friedel
//
//
//////////////////////////////////////////////////////////////////////////////

#ifndef PATCH_USED
#define PATCH_USED

//////////////////////////////////////////////////////////////////////////////

#include <assert.h>
#include <stdio.h>
#include <cmath>
#include <stdlib.h>
#include <iostream>
using namespace std;

//////////////////////////////////////////////////////////////////////////////

#define EPSILON0 1.0e-250
#define EPSILON1 1.0e-12
#define EPSILON2 1.0e-6
#define EPSILON3 1.0e-4
#ifndef M_PI
#define M_PI 3.14159265
#endif

//////////////////////////////////////////////////////////////////////////////

#define SIZE_N 3
#define SIZE_M 3
#define MAX_VALENCE 60

//////////////////////////////////////////////////////////////////////////////

class Patch;
typedef double (Patch::*F2D)(double,double);

//////////////////////////////////////////////////////////////////////////////

class Patch
{
 public:

  double     IS[2][2];
  double inv_IS[2][2];
  double    IIS[2][2];
  double   prod[2][2];

  double    D1S[3];
  double    D2S[3];
  double     NS[3];

  double  D1D1S[3];
  double  D1D2S[3];
  double  D2D1S[3];
  double  D2D2S[3];

  double  Gamma[2][2][2];

  double  det_IS, det_IIS, tmp;
  double  lambda1, lambda2;

  double  Hess_S_of_h[2][2];
  double  Grad_S_of_h[3];
  double  Domega;

  int IL_i,IL_j;
  Patch *ReferenceS;

//////////////////////////////////////////////////////////////////////////////

  inline  void   set_IS    (const double u, const double v);
  inline  void   set_inv_IS(void);
  inline  void   set_IIS   (void);
  inline  void   set_Gamma (void);
  inline  void   set_Hess_S_of_h(const double, const double,
				 const double, const double,
				 const double, const double,
				 const double);
  inline  void   set_Grad_S_of_h(const double, const double,
				 const double, const double);

//////////////////////////////////////////////////////////////////////////////

          double GaussInt (F2D,
			   const double,const double,
			   const double,const double,int);
          double NumInt   (F2D,
			   const double,const double,
			   const double,const double,
			   const double,const double,int);
          double Integrate(F2D,
			   const double,const double,
			   const double,const double);

  inline  double ExactBendingEnergyFunction(const double,const double);
  inline  double ExactBendingEnergyFunction_old(const double,const double);
  inline  double ExactMembraneEnergyFunction(const double,const double);
  virtual double Kp_function(const double,const double)=0;
  virtual double Km_function(const double,const double)=0;

//////////////////////////////////////////////////////////////////////////////

  virtual double evaluate_S(const double u, const double v, const int s)=0;

  virtual double evaluate_D1S(const double u, const double v, const int s)=0;
  virtual double evaluate_D2S(const double u, const double v, const int s)=0;

  virtual double evaluate_D1D1S(const double u, const double v, const int s)=0;
  virtual double evaluate_D1D2S(const double u, const double v, const int s)=0;
  virtual double evaluate_D2D1S(const double u, const double v, const int s)=0;
  virtual double evaluate_D2D2S(const double u, const double v, const int s)=0;

//////////////////////////////////////////////////////////////////////////////

  inline  void   set_principal_curvature(const double u, const double v);
  inline  double query_principal_curvature_1(void){return(lambda1);}
  inline  double query_principal_curvature_2(void){return(lambda2);}

  inline  double evaluate_mean_curvature(const double u, const double v);
  inline  double evaluate_Gaussian_curvature(const double u, const double v);

//////////////////////////////////////////////////////////////////////////////

  inline  double ExactBendingEnergy(void);
  inline  double ExactBendingEnergy_old(void);
  virtual void   Kp(Patch *S, double *K)=0;
  virtual double KpEnergy(void);
  //virtual double grad_KpEnergy(int coord, int i);
  // virtual double Hess_KpEnergy(int coord, int i); //****************
  // Gradient and Hessian of Bending Energy. Needed for minimizing..

  inline  double ExactMembraneEnergy(void);
  virtual void   Km(Patch *S, double *K)=0;
  virtual double KmEnergy(void);
  //virtual double grad_KmEnergy(int coord, int i);
  //virtual double Hess_KmEnergy(int coord, int i);

//////////////////////////////////////////////////////////////////////////////

  virtual void   print(void)=0;
  virtual void   export_IV(const char *fn,double,double,double);
  virtual void   export_IV_gaussian_coloring(const char *fn,double);
  virtual void   export_IV_mean_coloring(const char *fn,double);
  virtual void   export_IV_principal_coloring(const char *fn,double);

  //virtual void   matrix_mult( double *A, double *B, double *C, int I, int J, int K );
  //virtual void   copy_row(double *M_from, double *M_to, int row_from, int row_to, int M);
  //virtual void   copy_col(double *M_from, double *M_to, int col_from, int col_to, int M);
  //virtual void   convert_Ilja_Patrick(double *K, int M);

//////////////////////////////////////////////////////////////////////////////

  virtual void   update_parameters(double **data)
    {
      assert(data);
      cerr << "No update method implemented!"<<endl;
      exit(1);
    }

//////////////////////////////////////////////////////////////////////////////

  virtual ~Patch();
  Patch()
    {
      ReferenceS=NULL;
    }
};

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void Patch::set_IS(const double u, const double v)
{
  int i;

  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  for(i=0;i<3;i++)
    {
      D1S[i]  =evaluate_D1S(u,v,i);
      D2S[i]  =evaluate_D2S(u,v,i);

      D1D1S[i]=evaluate_D1D1S(u,v,i);
      D1D2S[i]=evaluate_D1D2S(u,v,i);
      D2D1S[i]=D1D2S[i];
      //D2D1S[i]=evaluate_D2D1S(u,v,i);
      //assert(D1D2S[i]==D2D1S[i]);
      D2D2S[i]=evaluate_D2D2S(u,v,i);
    }

  IS[0][0] = D1S[0]*D1S[0] + D1S[1]*D1S[1] + D1S[2]*D1S[2];
  IS[0][1] = D1S[0]*D2S[0] + D1S[1]*D2S[1] + D1S[2]*D2S[2];
  IS[1][0] = IS[0][1];
  IS[1][1] = D2S[0]*D2S[0] + D2S[1]*D2S[1] + D2S[2]*D2S[2];
}

//////////////////////////////////////////////////////////////////////////////
// first have to call set_IS()

void Patch::set_inv_IS(void)
{
  double inv_det_IS;

  det_IS    = IS[0][0]*IS[1][1]-IS[1][0]*IS[0][1];

  assert(fabs(det_IS)>EPSILON0);

  inv_det_IS= 1.0/det_IS;

  inv_IS[0][0] = inv_det_IS *  IS[1][1];
  inv_IS[0][1] = inv_det_IS * -IS[0][1];
  inv_IS[1][0] = inv_det_IS * -IS[1][0];
  inv_IS[1][1] = inv_det_IS *  IS[0][0];
}

//////////////////////////////////////////////////////////////////////////////
// first have to call set_inv_IS()

void Patch::set_IIS(void)
{
  NS[0] = D1S[1]*D2S[2]-D1S[2]*D2S[1];
  NS[1] = D1S[2]*D2S[0]-D1S[0]*D2S[2];
  NS[2] = D1S[0]*D2S[1]-D1S[1]*D2S[0];

  Domega= sqrt(NS[0]*NS[0] + NS[1]*NS[1] + NS[2]*NS[2]); // !!!!!!!!!!!!!!

  assert(fabs(Domega)>EPSILON0);

  tmp   = 1.0/Domega;

  NS[0] = tmp*NS[0];
  NS[1] = tmp*NS[1];
  NS[2] = tmp*NS[2];

  IIS[0][0] = D1D1S[0]*NS[0] + D1D1S[1]*NS[1] + D1D1S[2]*NS[2];
  IIS[0][1] = D1D2S[0]*NS[0] + D1D2S[1]*NS[1] + D1D2S[2]*NS[2];
  IIS[1][0] = IIS[0][1];
  IIS[1][1] = D2D2S[0]*NS[0] + D2D2S[1]*NS[1] + D2D2S[2]*NS[2];

  det_IIS = IIS[0][0]*IIS[1][1]-IIS[1][0]*IIS[0][1];
}

//////////////////////////////////////////////////////////////////////////////
// first have to call set_inv_IS()

void Patch::set_Gamma(void)
{
  int k;
  
  for(k=0;k<2;k++)
    {
      Gamma[0][0][k]=inv_IS[k][0]*(D1D1S[0]*D1S[0]+
				   D1D1S[1]*D1S[1]+
				   D1D1S[2]*D1S[2])+
	             inv_IS[k][1]*(D1D1S[0]*D2S[0]+
				   D1D1S[1]*D2S[1]+
				   D1D1S[2]*D2S[2]);

      Gamma[0][1][k]=inv_IS[k][0]*(D1D2S[0]*D1S[0]+
				   D1D2S[1]*D1S[1]+
				   D1D2S[2]*D1S[2])+
	             inv_IS[k][1]*(D1D2S[0]*D2S[0]+
				   D1D2S[1]*D2S[1]+
				   D1D2S[2]*D2S[2]);
      
      Gamma[1][0][k]=inv_IS[k][0]*(D2D1S[0]*D1S[0]+
				   D2D1S[1]*D1S[1]+
				   D2D1S[2]*D1S[2])+
                     inv_IS[k][1]*(D2D1S[0]*D2S[0]+
				   D2D1S[1]*D2S[1]+
				   D2D1S[2]*D2S[2]);

      Gamma[1][1][k]=inv_IS[k][0]*(D2D2S[0]*D1S[0]+
				   D2D2S[1]*D1S[1]+
				   D2D2S[2]*D1S[2])+
                     inv_IS[k][1]*(D2D2S[0]*D2S[0]+
				   D2D2S[1]*D2S[1]+
				   D2D2S[2]*D2S[2]);
    }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void Patch::set_principal_curvature(const double u, const double v)
{
  double   prod[2][2];
  double   a,b,d;

  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);
  
  set_IS(u,v);
  set_inv_IS();
  set_IIS();

  prod[0][0]=inv_IS[0][0]*IIS[0][0] + inv_IS[0][1]*IIS[1][0];
  prod[0][1]=inv_IS[0][0]*IIS[1][0] + inv_IS[0][1]*IIS[1][1];
  prod[1][0]=inv_IS[1][0]*IIS[0][0] + inv_IS[1][1]*IIS[1][0];
  prod[1][1]=inv_IS[1][0]*IIS[0][1] + inv_IS[1][1]*IIS[1][1];
  
  a = 0.5*prod[0][0]+0.5*prod[1][1];
  d = pow(prod[0][0],2)-
      2.0*prod[0][0]*prod[1][1]+
      pow(prod[1][1],2)+
      4.0*prod[0][1]*prod[1][0];

  assert(d>=-EPSILON1);
  if(d<0.0) d=0.0;

  b = 0.5*sqrt(d);
  lambda1=a+b;
  lambda2=a-b;
}

//////////////////////////////////////////////////////////////////////////////

double Patch::evaluate_mean_curvature(const double u, const double v)
{
  double   prod[2][2];

  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  set_IS(u,v);
  set_inv_IS();
  set_IIS();

    prod[0][0]=inv_IS[0][0]*IIS[0][0] + inv_IS[0][1]*IIS[1][0];
  //prod[0][1]=inv_IS[0][0]*IIS[1][0] + inv_IS[0][1]*IIS[1][1]; // not needed
  //prod[1][0]=inv_IS[1][0]*IIS[0][0] + inv_IS[1][1]*IIS[1][0]; // not needed
    prod[1][1]=inv_IS[1][0]*IIS[0][1] + inv_IS[1][1]*IIS[1][1];

  return(0.5*(prod[0][0]+prod[1][1]));
}
//////////////////////////////////////////////////////////////////////////////

double Patch::evaluate_Gaussian_curvature(const double u, const double v)
{
  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  set_IS(u,v);
  set_inv_IS();
  set_IIS();

  assert(fabs(det_IS)>EPSILON0);

  return(det_IIS/det_IS);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void Patch::set_Hess_S_of_h(const double u,     const double v, 
			    const double d1h,   const double d2h,
			    const double d1d1h, const double d1d2h, const double d2d2h)
{
  double d2d1h;

  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  d2d1h=d1d2h;

  set_IS(u,v);
  set_inv_IS();
  set_IIS();
  set_Gamma();

  // in tensor notation
  // 
  //  Hess_S_of_h[k][j]=inv_IS[k][l=0]*(djdlh - (di=1h*Gamma[j][l][i]+di=2h*Gamma[j][l][i]))+
  //                    inv_IS[k][l=1]*(djdlh - (di=1h*Gamma[j][l][i]+di=2h*Gamma[j][l][i]));
  //
  // unrolled
  //
  //  Hess_S_of_h[k=0][j=0]=inv_IS[k=0][l=0]*(dj=0dl=0h - (di=0h*Gamma[j=0][l=0][i=0]+di=1h*Gamma[j=0][l=0][i=1]))+
  //                        inv_IS[k=0][l=1]*(dj=0dl=1h - (di=0h*Gamma[j=0][l=1][i=0]+di=1h*Gamma[j=0][l=1][i=1]));
  //
  //  Hess_S_of_h[k=0][j=1]=inv_IS[k=0][l=0]*(dj=1dl=0h - (di=0h*Gamma[j=1][l=0][i=0]+di=1h*Gamma[j=1][l=0][i=1]))+
  //                        inv_IS[k=0][l=1]*(dj=1dl=1h - (di=0h*Gamma[j=1][l=1][i=0]+di=1h*Gamma[j=1][l=1][i=1]));
  //
  //  Hess_S_of_h[k=1][j=0]=inv_IS[k=1][l=0]*(dj=0dl=0h - (di=0h*Gamma[j=0][l=0][i=0]+di=1h*Gamma[j=0][l=0][i=1]))+
  //                        inv_IS[k=1][l=1]*(dj=0dl=1h - (di=0h*Gamma[j=0][l=1][i=0]+di=1h*Gamma[j=0][l=1][i=1]));
  // 
  //  Hess_S_of_h[k=1][j=1]=inv_IS[k=1][l=0]*(dj=1dl=0h - (di=0h*Gamma[j=1][l=0][i=0]+di=1h*Gamma[j=1][l=0][i=1]))+
  //                        inv_IS[k=1][l=1]*(dj=1dl=1h - (di=0h*Gamma[j=1][l=1][i=0]+di=1h*Gamma[j=1][l=1][i=1]));

  Hess_S_of_h[0][0]=inv_IS[0][0]*(d1d1h - (d1h*Gamma[0][0][0]+d2h*Gamma[0][0][1]))+
                    inv_IS[0][1]*(d1d2h - (d1h*Gamma[0][1][0]+d2h*Gamma[0][1][1]));

  Hess_S_of_h[0][1]=inv_IS[0][0]*(d2d1h - (d1h*Gamma[1][0][0]+d2h*Gamma[1][0][1]))+
                    inv_IS[0][1]*(d2d2h - (d1h*Gamma[1][1][0]+d2h*Gamma[1][1][1])); 

  // ********
  // not symmetrical ???????
  // ********
  // formulas for Hess_S_of_h[0][1] and Hess_S_of_h[1][0] are correct as far as I can see (paper)

  Hess_S_of_h[1][0]=inv_IS[1][0]*(d1d1h - (d1h*Gamma[0][0][0]+d2h*Gamma[0][0][1]))+
                    inv_IS[1][1]*(d1d2h - (d1h*Gamma[0][1][0]+d2h*Gamma[0][1][1]));

  Hess_S_of_h[1][1]=inv_IS[1][0]*(d2d1h - (d1h*Gamma[1][0][0]+d2h*Gamma[1][0][1]))+
                    inv_IS[1][1]*(d2d2h - (d1h*Gamma[1][1][0]+d2h*Gamma[1][1][1]));
}

//////////////////////////////////////////////////////////////////////////////

void Patch::set_Grad_S_of_h(const double u,     const double v, 
			    const double d1h,   const double d2h)
{
  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  set_IS(u,v);
  set_inv_IS();

  Grad_S_of_h[0]=inv_IS[0][0]*d1h*D1S[0]+
                 inv_IS[0][1]*d2h*D1S[0]+
                 inv_IS[1][0]*d1h*D2S[0]+
                 inv_IS[1][1]*d2h*D2S[0];

  Grad_S_of_h[1]=inv_IS[0][0]*d1h*D1S[1]+
                 inv_IS[0][1]*d2h*D1S[1]+
                 inv_IS[1][0]*d1h*D2S[1]+
                 inv_IS[1][1]*d2h*D2S[1];

  Grad_S_of_h[2]=inv_IS[0][0]*d1h*D1S[2]+
                 inv_IS[0][1]*d2h*D1S[2]+
                 inv_IS[1][0]*d1h*D2S[2]+
                 inv_IS[1][1]*d2h*D2S[2];
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

double Patch::ExactBendingEnergyFunction(const double u, const double v)
{
  double trace_HH1, trace_HH2, trace_HH3;
  double d1h, d2h, d1d1h, d1d2h, d2d2h;

  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);
  
  set_IS(u,v);
  set_inv_IS();
  set_IIS();

  d1h  = evaluate_D1S(u,v,0);
  d2h  = evaluate_D2S(u,v,0);
  d1d1h= evaluate_D1D1S(u,v,0);
  d1d2h= evaluate_D1D2S(u,v,0);
  d2d2h= evaluate_D2D2S(u,v,0);

  set_Hess_S_of_h(u,v, 
		  d1h,d2h, 
		  d1d1h,d1d2h,d2d2h);
  trace_HH1 =     pow(Hess_S_of_h[0][0],2.0) +
              2.0*Hess_S_of_h[0][1]*Hess_S_of_h[1][0] +
                  pow(Hess_S_of_h[1][1],2.0);

  d1h  = evaluate_D1S(u,v,1);
  d2h  = evaluate_D2S(u,v,1);
  d1d1h= evaluate_D1D1S(u,v,1);
  d1d2h= evaluate_D1D2S(u,v,1);
  d2d2h= evaluate_D2D2S(u,v,1);

  set_Hess_S_of_h(u,v, 
		  d1h,d2h, 
		  d1d1h,d1d2h,d2d2h);
  trace_HH2 =     pow(Hess_S_of_h[0][0],2.0) +
              2.0*Hess_S_of_h[0][1]*Hess_S_of_h[1][0] +
                  pow(Hess_S_of_h[1][1],2.0);

  d1h  = evaluate_D1S(u,v,2);
  d2h  = evaluate_D2S(u,v,2);
  d1d1h= evaluate_D1D1S(u,v,2);
  d1d2h= evaluate_D1D2S(u,v,2);
  d2d2h= evaluate_D2D2S(u,v,2);

  set_Hess_S_of_h(u,v, 
		  d1h,d2h, 
		  d1d1h,d1d2h,d2d2h);
  trace_HH3 =     pow(Hess_S_of_h[0][0],2.0) +
              2.0*Hess_S_of_h[0][1]*Hess_S_of_h[1][0] +
                  pow(Hess_S_of_h[1][1],2.0);

  return((trace_HH1 + trace_HH2 + trace_HH3)*Domega);
}
//////////////////////////////////////////////////////////////////////////////

double Patch::ExactBendingEnergyFunction_old(const double u, const double v)
{
  double tmp;

  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);
  
  set_principal_curvature(u,v);
  tmp=(lambda1*lambda1+lambda2*lambda2)*Domega;
  return(tmp);
}

//////////////////////////////////////////////////////////////////////////////

double Patch::ExactBendingEnergy(void)
{
  double E;

  E=Integrate((F2D)&Patch::ExactBendingEnergyFunction,
  	      0.0, 1.0, 
  	      0.0, 1.0);
  return(E);
}

//////////////////////////////////////////////////////////////////////////////

double Patch::ExactBendingEnergy_old(void)
{
  double E;

  E=Integrate((F2D)&Patch::ExactBendingEnergyFunction_old,
  	      0.0, 1.0, 
  	      0.0, 1.0);
  return(E);
}

//////////////////////////////////////////////////////////////////////////////

double Patch::ExactMembraneEnergyFunction(const double u, const double v)
{
  int    i;
  double tmp;
  double D1F[3];
  double D2F[3];

  assert(u>=0.0 && u<=1.0);
  assert(v>=0.0 && v<=1.0);

  for(i=0;i<3;i++)
    {
      D1F[i]=evaluate_D1S(u,v,i);
      D2F[i]=evaluate_D2S(u,v,i);
    }

  tmp=0.0;

  for(i=0;i<3;i++)
    {
      set_Grad_S_of_h(u,v,D1F[i],D2F[i]);
      tmp+=Grad_S_of_h[0]*Grad_S_of_h[0]+
	   Grad_S_of_h[1]*Grad_S_of_h[1]+
	   Grad_S_of_h[2]*Grad_S_of_h[2];
    }
  set_IIS();                        // only Domega needed
  tmp=tmp*Domega;

  return(tmp);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

double Patch::ExactMembraneEnergy(void)
{
  double E;

  E=Integrate((F2D)&Patch::ExactMembraneEnergyFunction,
 	      0.0, 1.0, 
  	      0.0, 1.0);

  return(E);  
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#endif
